
# $Id: remote.py,v 1.25 2004/09/26 19:58:01 cran Exp $
#
# Copyright (c) 2002, 2003, 2004 Sebastian Stark
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR SEBASTIAN STARK
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""Various remote methods for tentakel

Currently, only the "ssh" method is implemented."""

import error
import sys
import commands
import threading, Queue
import tpg
import time

class FormatString(tpg.Parser):
	r"""

	token escape	: '\\[\\nt]'	str ;
	token fmtchar	: '%[%dost]'	str ;
	token char	: '.'		str ;

	START/e -> FORMAT/e ;

	FORMAT/f ->
					$ f = ""
			( escape/e	$ f = f + self.getEscape(e)
			| fmtchar/fc	$ f = f + self.getSpecialChar(fc)
			| char/c	$ f = f + c
			)*
	;
	"""

	def __init__(self):
		super(FormatString, self).__init__()
		self.map = {r"%%": "%"}
	
	def getMap(self):
		return self.map
	
	def setMap(self, userMap):
		self.map.update(userMap)
	
	formatMap = property(getMap, setMap, doc="current format character mapping")
	
	def getEscape(self, e):
		"Return a dictionary which maps escape strings to literals"
		return {
			r"\\": r"\\",
			r"\n": "\n",
			r"\t": "\t"
		}[e]

	def getSpecialChar(self,s):
		return self.formatMap[s]


class RemoteCommand(threading.Thread):
	"""Generic remote execution class

	Specific remote command classes should inherit from this class
	and define a _rexec() method that executes the command and
	puts the result into the resultQueue.

	The __init__ method can be overridden if special processing of
	the params parameter should be done. In that case,
	RemoteCommand.__init__(self, destination, params) should
	be called at the end of __init__.

	The _rexec() method should measure the time it needs to run and
	set duration accordingly.
	"""

	def __init__(self, destination, params):
		threading.Thread.__init__(self)
		self.setDaemon(1)
		self.duration = 0.0
		self.destination = destination
		self.commandQueue = Queue.Queue()
		self.resultQueue = Queue.Queue()
		self.start()

	def execute(self, command):
		"Execute a command in this thread"

		self.commandQueue.put(command)

	def run(self):
		while 1:
			self._rexec()

def remoteCommandFactory(destination, params):
	"""Depending in the method, instantiate a corresponding
	RemoteCommand derived object and return it"""
	
	method = params['method']
	if method == 'ssh':
		return SSHRemoteCommand(destination, params)
	elif method == 'rsh':
		return RSHRemoteCommand(destination, params)
	else:
		error.err('Method not implemented: "%s"' % method)


class SSHRemoteCommand(RemoteCommand):
	"SSH remote execution class"

	def __init__(self, destination, params):
		self.sshpath = params['ssh_path']
		self.user = params['user']
		RemoteCommand.__init__(self, destination, params)

	def _rexec(self):
		s = '%s %s@%s "%s"' % (self.sshpath, self.user, self.destination, self.commandQueue.get())
		t1 = time.time()
		status, output = commands.getstatusoutput(s)
		self.duration = time.time() - t1
		# shift 8 bits right to strip signal number from status
		self.resultQueue.put((status >> 8, output))


class RSHRemoteCommand(RemoteCommand):
	"RSH remote execution class"

	def __init__(self, destination, params):
		self.rshpath = params['rsh_path']
		self.user = params['user']
		RemoteCommand.__init__(self, destination, params)

	def _rexec(self):
		s = '%s -l %s %s "%s"' % (self.rshpath, self.user, self.destination, self.commandQueue.get())
		t1 = time.time()
		status, output = commands.getstatusoutput(s)
		self.duration = time.time() - t1
		# shift 8 bits right to strip signal number from status
		self.resultQueue.put((status >> 8, output))


class RemoteCollator(object):
	"""This class is meant to hold RemoteCommand instances each
	of which implements a specific way too execute a command on
	a remote host"""

	def __init__(self, conf, groupName):
		self.clear()
		self.useConf(conf, groupName)
		self.formatter = FormatString()

	def clear(self):
		"Empty the list of contained remoteobjects"
		self.remoteobjects = []

	def useConf(self, conf, groupName):
		"""Load the specified group from configuration object conf
		and add RemoteCommand objects for each contained host"""
		save = self
		self.clear()
	        try:
	                for destination, params in conf.getGroupMembers(groupName):
	                        self.add(remoteCommandFactory(destination, params))
			self.format = conf.getParam("format", group=groupName)
	        except KeyError:
			self = save
	                error.warn("unknown group: '%s'" % groupName)

	def getDestinations(self):
		"Return expanded list of hosts"
		return [x.destination for x in self.remoteobjects]

	def add(self,obj):
		"Add a RemoteObject"
		if isinstance(obj,RemoteCommand):
			self.remoteobjects.append(obj)
		else:
			pass
	
	def remove(self,obj):
		"Remove a RemoteObject"
		self.remoteobjects.remove(obj)

	def expandFormat(self, map=None):
		"""Apply a format mapping to the format string

		Outputs the format with formatting expressions replaced by
		values taken from map. The map must contain translations for
		the formatting expressions. For example:

		  map = { r"%d": "something" }

		"""

		self.formatter.formatMap = map
		return self.formatter(self.format)

	def execAll(self, command):
		"Execute command on all remote objects"

		# TODO: think about this a bit
		for obj in self.remoteobjects:
			obj.execute(command)

	def displayAll(self):
		"Display the results of all remote objects"

		# TODO: output results when they're ready, not later
		for obj in self.remoteobjects:
			status, output = obj.resultQueue.get()
			resultMap = {
					r"%d": obj.destination,
					r"%t": str(round(obj.duration, 2)),
					r"%o": output,
					r"%s": str(status)
				}
			sys.stdout.write(self.expandFormat(resultMap))


if __name__ == "__main__":

	failures = 0
	print "self testing..."

	print "### instantiate SSHRemoteCommand:"
	import pwd, os
	user = pwd.getpwuid(os.geteuid())[0]
	r = remoteCommandFactory('localhost', {'method': 'ssh', 'user': user, 'ssh_path': '/usr/bin/ssh'})
	if isinstance(r, SSHRemoteCommand):
		print "OK"
	else:
		print "-> failed <-"
		failures += 1

	print "### execute command:"
	v = (0, "test123")
	r.execute("echo test123")
	res = r.resultQueue.get()
	if res == v:
		print "OK"
	else:
		print "-> failed <-"
		print "read", res, "but should be", v
		failures += 1

	print "### instantiate RSHRemoteCommand:"
	import pwd, os
	user = pwd.getpwuid(os.geteuid())[0]
	r = remoteCommandFactory('localhost', {'method': 'rsh', 'user': user, 'rsh_path': '/usr/bin/rsh'})
	if isinstance(r, RSHRemoteCommand):
		print "OK"
	else:
		print "-> failed <-"
		failures += 1

	print "### execute command:"
	v = (0, "test123")
	r.execute("echo test123")
	res = r.resultQueue.get()
	if res == v:
		print "OK"
	else:
		print "-> failed <-"
		print "read", res, "but should be", v
		failures += 1


	if failures:
		print "self test: encountered", failures, "failures."
		sys.exit(1)
	else:
		print "self test: no failures."
		sys.exit(0)


